from os import path

import boto3

from johnsnowlabs.utils.s3_utils import parse_s3_url, upload_file_to_s3


def upload_notebook_to_s3(
    boto_session: boto3.Session,
    workspace_storage_s3_url: str,
    editor_id: str,
    path_to_notebook: str,
) -> str:
    """Uploads a notebook to S3 and returns its s3 url
    :param workspace_storage_s3_url: S3 path to workspace storage
    :param editor_id: EMR editor id
    :param path_to_notebook: Local Path to notebook
    """
    if not boto_session:
        boto_session = boto3.Session()
    # Upload local notebook to S3
    if not path.exists(path_to_notebook):
        raise FileNotFoundError(f"Notebook {path_to_notebook} not found")
    notebook_name = path.basename(path_to_notebook)

    bucket, key = parse_s3_url(workspace_storage_s3_url)

    file_name_key = f"{key}/{editor_id}/{notebook_name}"

    return upload_file_to_s3(
        boto_session=boto_session,
        file_path=path_to_notebook,
        bucket=bucket,
        file_name=file_name_key,
    )


def run_local_notebook(
    boto_session: boto3.Session,
    workspace_storage_s3_url: str,
    cluster_id: str,
    editor_id: str,
    path_to_notebook: str,
):
    """Runs an EMR notebook using boto3"""
    if not boto_session:
        boto_session = boto3.Session()

    emr_client = boto_session.client("emr")
    upload_notebook_to_s3(
        boto_session=boto_session,
        workspace_storage_s3_url=workspace_storage_s3_url,
        editor_id=editor_id,
        path_to_notebook=path_to_notebook,
    )

    notebook_name = path.basename(path_to_notebook)

    response = emr_client.start_notebook_execution(
        EditorId=editor_id,
        RelativePath=notebook_name,
        ExecutionEngine={"Id": cluster_id},
        ServiceRole="EMR_Studio_Service_Role",
    )
    return response["NotebookExecutionId"]



import mimetypes
import os
import subprocess
from pathlib import Path
from typing import Optional

import requests

from johnsnowlabs import settings
from johnsnowlabs.py_models.jsl_secrets import JslSecrets
from johnsnowlabs.utils.enums import JvmHardwareTarget
from johnsnowlabs.utils.env_utils import get_folder_of_func
from johnsnowlabs.utils.py_process import run_cmd_and_check_succ


def check_image_exist(image_name: str) -> bool:
    cmd = f"docker image inspect {image_name}"
    return run_cmd_and_check_succ(
        [cmd], shell=True, raise_on_fail=False, use_code=True, log=False, log_outputs=False
    )


def check_container_exist(container_name: str) -> bool:
    cmd = f"docker container inspect {container_name}"
    return run_cmd_and_check_succ(
        [cmd], shell=True, raise_on_fail=False, use_code=True, log=False, log_outputs=False
    )


def _destroy_container(container_name: str = None):
    container_name = (
        settings.docker_container_name if container_name is None else container_name
    )

    if check_container_exist(container_name):
        stop_cmd = f"docker container stop {container_name}"
        rm_cmd = f"docker container rm -f {container_name}"
        run_cmd_and_check_succ(
            [stop_cmd], shell=True, raise_on_fail=False, use_code=True,log_outputs=False
        )
        run_cmd_and_check_succ([rm_cmd], shell=True, raise_on_fail=False, use_code=True,log_outputs=False)
        print(f"Container '{container_name}' destroyed.")
    else:
        print(f"Container '{container_name}' does not exist.")


def _destroy_image(image_name: str = None):
    image_name = settings.docker_image_name if image_name is None else image_name
    if check_image_exist(image_name):
        rm_cmd = f"docker image rm -f {image_name}"
        run_cmd_and_check_succ([rm_cmd], shell=True, raise_on_fail=False, use_code=True, log_outputs=False)
        print(f"Image '{image_name}' destroyed.")
    else:
        print(f"Tried to destroy image '{image_name}', but does not exist.")


def is_docker_installed():
    try:
        # Command to check Docker version
        command = "asd asd docker dasdasd --version"

        # Execute the command
        result = subprocess.run(
            command,
            shell=True,
            check=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )

        # Decode the output
        output = result.stdout.decode().strip()

        if "Docker" in output:
            return True
        else:
            return False

    except subprocess.CalledProcessError as e:
        return False


def generate_dockerfile(
        # Install parameters
        model: str,
        secrets: JslSecrets,
        visual: bool = False,
        nlp: bool = True,
        hardware_platform: str = JvmHardwareTarget.cpu.value,
):
    """
    Generate a Dockerfile for a specific model and configuration.

    :param model: NLU Reference to the model to be used.
    :param secrets: JslSecrets object
    :param visual: Flag to include visual features.
    :param nlp: Flag to include medical NLP features.
    :param hardware_platform: Target hardware platform (e.g., 'cpu').
    """
    if not model:
        raise Exception("model may not me None", model)

    build_folder = os.path.join(get_folder_of_func(_destroy_container), "build")
    base_docker_file_path = os.path.join(build_folder, "base_dockerfile")
    generated_docker_file_path = os.path.join(build_folder, "generated_dockerfile")

    provided_license = None
    if secrets.OCR_LICENSE:
        provided_license = secrets.OCR_LICENSE

    if secrets.HC_LICENSE:
        provided_license = secrets.HC_LICENSE

    env_vars = [
        f'ENV HARDWARE_TARGET="{JvmHardwareTarget(hardware_platform).value}"',
        f'ENV MODEL_TO_LOAD="{model}"',
    ]

    if provided_license:
        env_vars.append(f'ENV JOHNSNOWLABS_LICENSE="{provided_license}"')

    if provided_license:
        env_vars.append(f'ENV JOHNSNOWLABS_LICENSE="{provided_license}"')
    if secrets.HC_SECRET and nlp:
        env_vars.append(f'ENV MEDICAL_SECRET="{secrets.HC_SECRET}"')
    if secrets.OCR_SECRET and visual:
        env_vars.append(f'ENV VISUAL_SECRET="{secrets.OCR_SECRET}"')
    if secrets.AWS_ACCESS_KEY_ID:
        env_vars.append(
            f'ENV JOHNSNOWLABS_AWS_ACCESS_KEY_ID="{secrets.AWS_ACCESS_KEY_ID}"'
        )
    if secrets.AWS_SECRET_ACCESS_KEY:
        env_vars.append(
            f'ENV JOHNSNOWLABS_AWS_SECRET_ACCESS_KEY="{secrets.AWS_SECRET_ACCESS_KEY}"'
        )

    with open(base_docker_file_path, "r", encoding="utf-8") as file:
        docker_code = file.read()

    docker_code = insert_strings(3, env_vars, docker_code)

    with open(generated_docker_file_path, "w", encoding="utf-8") as file:
        file.write(docker_code)


def insert_strings(index, inserts, start_string):
    """
    Insert multiple strings into a starting string at a specified index.

    :param index: Position in the starting string where the inserts begin.
    :param inserts: List of strings to insert.
    :param start_string: The initial string into which the inserts are made.
    :return: Modified string with inserts.
    """

    lines = start_string.split("\n")
    for string in inserts:
        lines.insert(index + 1, string)
        index += 1
    return "\n".join(lines)


def build_image(
        preloaded_model: str,  # nlu ref, nlp_ref or lcoally stored model
        image_name=None,
        rebuild=False,
        use_cache=True,
        # Install parameters
        # -- JSL-Auth Flows --
        # Browser Auth
        browser_login: bool = False,
        # JWT Token Auth
        access_token: Optional[str] = None,
        # JSON file Auth
        json_license_path: Optional[str] = None,
        # Manual License specification Auth
        med_license: Optional[str] = None,
        enterprise_nlp_secret: Optional[str] = None,
        ocr_secret: Optional[str] = None,
        ocr_license: Optional[str] = None,
        fin_license: Optional[str] = None,
        leg_license: Optional[str] = None,
        aws_access_key: Optional[str] = None,
        aws_key_id: Optional[str] = None,
        # Download Params
        nlp: bool = True,
        visual: bool = False,
        # License usage & Caching
        local_license_number: int = 0,
        remote_license_number: int = 0,
        store_in_jsl_home: bool = True,
        # Install File Types
        hardware_platform: str = JvmHardwareTarget.cpu.value,
):
    """
    Build a Docker image with specified parameters.

    :param preloaded_model: Reference to the preloaded model.
    :param image_name: Name of the Docker image. If None, uses default from settings.
    :param rebuild: Flag to destroy existing image and rebuild
    :param use_cache: Flag to use cache during the build.
    :param json_license_path: Path to the JSON license file.
    :param access_token: Access token for authentication.
    :param visual: Flag to include visual features.
    :param nlp: Flag to include NLP features.
    :param hardware_platform: Target hardware platform (e.g., 'cpu').
    """
    image_name = settings.docker_image_name if image_name is None else image_name
    if rebuild:
        _destroy_image(image_name)

    build_folder = os.path.join(get_folder_of_func(_destroy_container), "build")
    cmd = f"cd {build_folder} && docker build -f generated_dockerfile . -t {image_name}"
    if not use_cache:
        cmd += " --no-cache"
    # if browser_login or any([access_token,json_license_path, med_license,enterprise_nlp_secret, ocr_secret,ocr_license, fin_license,leg_license,aws_access_key,aws_key_id]):
    secrets: JslSecrets = JslSecrets.build_or_try_find_secrets(
        browser_login=browser_login,
        access_token=access_token,
        secrets_file=json_license_path,
        hc_license=med_license,
        hc_secret=enterprise_nlp_secret,
        ocr_secret=ocr_secret,
        ocr_license=ocr_license,
        fin_license=fin_license,
        leg_license=leg_license,
        aws_access_key=aws_access_key,
        aws_key_id=aws_key_id,
        local_license_number=local_license_number,
        remote_license_number=remote_license_number,
        store_in_jsl_home=store_in_jsl_home,
        return_empty_secrets_if_none_found=True,
    )

    generate_dockerfile(
        model=preloaded_model,
        secrets=secrets,
        visual=visual,
        nlp=nlp,
        hardware_platform=hardware_platform,
    )
    run_cmd_and_check_succ([cmd], shell=True, raise_on_fail=True, use_code=True, log_outputs=False)


def run_container_cmd(container_name=None, image_name=None, destroy_container=False):
    """
    Run a command in a Docker container.

    :param container_name: Name of the Docker container. If None, uses default from settings.
    :param image_name: Name of the Docker image. If None, uses default from settings.
    :param destroy_container: Flag to destroy the container before running the command.
    """

    image_name = settings.docker_image_name if image_name is None else image_name
    container_name = (
        settings.docker_container_name if container_name is None else container_name
    )
    if destroy_container:
        _destroy_container(container_name)

    # Command to run the container, add tail -f /dev/null to keep the container running
    cmd = f"docker run --name {container_name} -d {image_name} tail -f /dev/null"

    run_cmd_and_check_succ([cmd], shell=True, raise_on_fail=True, use_code=True)


def serve_container(
        container_name=None, image_name=None, destroy_container=False, host_port=8000
):
    """
    Serve a Docker container on a specified host port.

    :param container_name: Name of the Docker container. If None, uses default from settings.
    :param image_name: Name of the Docker image. If None, uses default from settings.
    :param destroy_container: Flag to destroy the container before serving.
    :param host_port: Host port to bind the container's service port.
    """
    image_name = settings.docker_image_name if image_name is None else image_name
    container_name = (
        settings.docker_container_name if container_name is None else container_name
    )

    if destroy_container:
        _destroy_container(container_name)

    # Command to run the container
    # Map the container's port 80 to the host's specified port
    cmd = f"docker run --name {container_name} -p {host_port}:80 -d {image_name}"

    run_cmd_and_check_succ([cmd], shell=True, raise_on_fail=True, use_code=True)


def send_file_to_server(file_path, port):
    """Send a file to the server using multipart/form-data."""
    try:
        # Fallback to binary stream type if MIME type is None
        mime_type = mimetypes.guess_type(file_path)[0] or 'application/octet-stream'
        files = {'file': (Path(file_path).name, open(file_path, 'rb'), mime_type)}
        response = requests.post(f"http://localhost:{port}/predict_file", files=files)
        response.raise_for_status()
        return response.json()  # Assuming your server responds with JSON
    except Exception as e:
        print(f"Error sending file to server: {e}")
        return None


def download_file(url):
    """Download a file from a URL to the current working directory if it doesn't already exist."""
    local_filename = url.split('/')[-1]  # Extract the filename from the URL
    local_filepath = Path(local_filename)
    if not local_filepath.is_file():  # Check if the file already exists
        with requests.get(url, stream=True) as r:
            r.raise_for_status()
            with open(local_filename, 'wb') as f:
                for chunk in r.iter_content(chunk_size=8192):
                    f.write(chunk)
        print(f"File '{local_filename}' downloaded.")
    else:
        print(f"File '{local_filename}' already exists.")
    return local_filename


def health_check_ocr_container(port, file_url):
    if file_url.startswith('http://') or file_url.startswith('https://'):
        response = send_file_to_server(download_file(file_url), port)
    elif os.path.exists(file_url):
        response = send_file_to_server(file_url, port)
    else:
        raise ValueError(f'Invalid path {file_url}')
    print(response)
    return True


def check_local_endpoint_health(port):
    url = f"http://localhost:{port}/predict_batch"
    params = {
        "text": ["Your text that you want to predict with the model goes here",
                 'It can also be a list of strings'],
    }
    response = requests.post(url, params=params, headers={"accept": "application/json"})
    print('BATCH PREDICTION UNPARAMETERIZED:')
    print(response.json())
    response.raise_for_status()
    params = {
        "text": "Your text that you want to predict with the model goes here",
    }
    url = f"http://localhost:{port}/predict"
    response = requests.get(url, params=params, headers={"accept": "application/json"})
    response.raise_for_status()
    print('STRING PREDICTION UNPARAMETERIZED:')
    print(response.json())

    params = {
        "text": "Your text that you want to predict with the model goes here",
        "output_level": "document",
        "positions": "false",
        "metadata": "false",
        "drop_irrelevant_cols": "false",
        "get_embeddings": "false",
        "keep_stranger_features": "true",
    }
    response = requests.get(url, params=params, headers={"accept": "application/json"})
    response.raise_for_status()
    print('STRING PREDICTION PARAMETERIZED:')
    print(response.json())

    print('NLP Endpoint ok!')
    return True



# Embedder Node compatible with haystack framework
import os
import sys
from pathlib import Path
from typing import List, Optional, Union, Dict, Any, Tuple

import numpy as np
from haystack import BaseComponent
from haystack.document_stores import BaseDocumentStore
from haystack.nodes.retriever import EmbeddingRetriever
from haystack.nodes.retriever._base_embedding_encoder import _BaseEmbeddingEncoder
from haystack.schema import Document, MultiLabel
from tqdm.auto import tqdm

from johnsnowlabs.frameworks.embedding_retrieval.utils import (
    get_docsplitter_pipe,
    get_doc_splitter_internal_pipe,
)


class _JohnsnowlabsEmbeddingEncoder(_BaseEmbeddingEncoder):
    def __init__(self, retriever: "EmbeddingRetriever"):
        # 1) Check imports
        try:
            from johnsnowlabs import nlp
            from nlu.pipe.pipeline import NLUPipeline
        except ImportError as exc:
            raise ImportError(
                "Could not import johnsnowlabs python package. "
                "Please install it with `pip install johnsnowlabs`."
            ) from exc

        # 2) Start a Spark Session
        try:
            os.environ["PYSPARK_PYTHON"] = sys.executable
            os.environ["PYSPARK_DRIVER_PYTHON"] = sys.executable
            nlp.start(hardware_target="gpu" if retriever.use_gpu else "cpu")
        except Exception as exc:
            raise Exception("Failure starting Spark Session") from exc
            # 3) Load the model
        try:
            self.embedding_model = nlp.load(retriever.embedding_model)
        except Exception as exc:
            raise Exception("Failure loading model") from exc

    def embed(self, texts: Union[List[str], str]) -> np.ndarray:
        return np.asarray(
            self.embedding_model.predict_embeds(texts),
            dtype=float,
        )

    def embed_queries(self, queries: List[str]) -> np.ndarray:
        return self.embed(queries)

    def embed_documents(self, docs: List[Document]) -> np.ndarray:
        return self.embed([d.content for d in docs])

    def train(
        **kwargs,
    ):
        raise NotImplementedError("Training not supported")

    def save(self, save_dir: Union[Path, str]):
        raise NotImplementedError("Saving not supported")


class JohnSnowLabsHaystackEmbedder(EmbeddingRetriever):
    def __init__(
        self,
        embedding_model: str,
        use_gpu: bool = False,
        document_store: Optional[BaseDocumentStore] = None,
        **kwargs,
    ):
        inject()
        kwargs["model_format"] = "johnsnowlabs"
        super().__init__(
            embedding_model=embedding_model,
            document_store=document_store,
            use_gpu=use_gpu,
            **kwargs,
        )


class JohnSnowLabsHaystackDocumentSplitter(BaseComponent):  # BasePreProcessor
    outgoing_edges = 1

    def __init__(
        self,
        chunk_size=500,
        chunk_overlap=0,
        split_patterns=["\n\n", "\n", " ", ""],
        explode_splits=True,
        keep_separators=True,
        patterns_are_regex=False,
        trim_whitespace=True,
        ##### OLD PARAMS #####
        progress_bar: bool = True,
    ):
        """
        :param chunk_size: Size of each chunk of text.
        :param chunk_overlap: Length of the overlap between text chunks , by default `0`.
        :param split_patterns: Patterns to separate the text by in decreasing priority , by default `["\n\n", "\n", " ", ""]`.
        :param explode_splits: Whether to explode split chunks to separate rows , by default `True`.
        :param keep_separators: Whether to keep the separators in the final result , by default `True`.
        :param patterns_are_regex: Whether to interpret the split patterns as regular expressions , by default `False`.
        :param trim_whitespace: Whether to trim whitespaces of extracted chunks , by default `True`.
        :param progress_bar: Whether to show a progress bar.
        """
        super().__init__()
        self.progress_bar = progress_bar
        self.pipe = get_docsplitter_pipe(
            chunk_size=chunk_size,
            chunk_overlap=chunk_overlap,
            split_patterns=split_patterns,
            explode_splits=explode_splits,
            keep_separators=keep_separators,
            patterns_are_regex=patterns_are_regex,
            trim_whitespace=trim_whitespace,
        )

    def process(
        self,
        documents: Union[dict, Document, List[Union[dict, Document]]] = None,
    ) -> List[Document]:
        """
        Perform document cleaning and splitting. Can take a single document or a list of documents as input and returns a list of documents.
        """
        if isinstance(documents, (Document, dict)):
            ret = self._process_single(document=documents)
        elif isinstance(documents, list):
            ret = self._process_batch(documents=list(documents))
        elif documents is None:
            ret = []
        else:
            raise Exception(
                f"documents provided to PreProcessor.prepreprocess() is not of type list nor Document or None but is {type(documents)} "
            )

        return ret

    def _process_single(
        self,
        document: Union[dict, Document],
    ) -> List[Document]:
        return self.split(document=self.clean(document=document))

    def _process_batch(self, documents: List[Union[dict, Document]]) -> List[Document]:
        nested_docs = [
            self._process_single(d)
            for d in tqdm(
                documents,
                disable=not self.progress_bar,
                desc="Preprocessing",
                unit="docs",
            )
        ]
        return [d for x in nested_docs for d in x]

    def clean(
        self,
        document: Union[dict, Document],
        # not implemented
        clean_whitespace: bool = None,
        clean_header_footer: bool = None,
        clean_empty_lines: bool = None,
        remove_substrings: Optional[List[str]] = None,
    ) -> Document:
        """
        Stub for Feature cleaning procedure
        """
        return document

    def split(
        self,
        document: Union[dict, Document],
    ) -> List[Document]:
        """Perform document splitting on a single document. This method can split on different units, at different lengths,
        with different strides. It can also respect sentence boundaries. Its exact functionality is defined by
        the parameters passed into PreProcessor.__init__(). Takes a single document as input and returns a list of documents.
        """
        # Todo pass params dynamically?
        text = document.content
        texts = [split for split in self.pipe.annotate(text)["splits"]]
        return [
            Document(id_hash_keys=["content", "meta"], content=text, meta=document.meta)
            for i, text in enumerate(texts)
        ]

    def run(
        self,
        query: Optional[str] = None,
        file_paths: Optional[List[str]] = None,
        labels: Optional[MultiLabel] = None,
        documents: Optional[List[Document]] = None,
        meta: Optional[dict] = None,
    ) -> Tuple[Dict, str]:
        """
        Method that will be executed when the node in the graph is called.

        The argument that are passed can vary between different types of nodes
        (e.g. retriever nodes expect different args than a reader node)

        See an example for an implementation in haystack/reader/base/BaseReader.py
        :return:
        """
        # result = {"documents": d.to_dict() for d in self.process(documents)}
        result = {"documents": d for d in self.process(documents)}
        return result, "output_1"

    def run_batch(
        self,
        queries: Optional[Union[str, List[str]]] = None,
        file_paths: Optional[List[str]] = None,
        labels: Optional[Union[MultiLabel, List[MultiLabel]]] = None,
        documents: Optional[Union[List[Document], List[List[Document]]]] = None,
        meta: Optional[Union[Dict[str, Any], List[Dict[str, Any]]]] = None,
        params: Optional[dict] = None,
        debug: Optional[bool] = None,
    ):
        result = {"documents": d for d in self.process(documents)}
        return result, "output_1"


class JohnSnowLabsHaystackMedicalDocumentSplitter(
    JohnSnowLabsHaystackDocumentSplitter
):  # BasePreProcessor
    outgoing_edges = 1

    def __init__(
        self,
        # Splitter params
        split_mode: str = "recursive",
        chunk_size: int = 500,
        chunk_overlap: int = 0,
        split_patterns: list = None,    # different value for split modes
        explode_splits: bool = True,
        keep_separators: bool = True,
        patterns_are_regex: bool = False,
        trim_whitespace: bool = True,
        sentence_awareness: bool = False,
        max_length: int = None,         # different value for split modes
        custom_bounds_strategy: str = "prepend",
        case_sensitive: bool = False,
        meta_data_fields: list = [],
        enable_sentence_increment: bool = False,
        # Tokenizer Params
        tokenizer_target_pattern: str = None,
        tokenizer_prefix_pattern: str = None,
        tokenizer_suffix_pattern: str = None,
        tokenizer_infix_patterns: list = None,
        tokenizer_exceptions: list = None,
        tokenizer_exceptions_path: str = None,
        tokenizer_case_sensitive_exceptions: bool = None,
        tokenizer_context_chars: list = None,
        tokenizer_split_pattern: str = None,
        tokenizer_split_chars: list = None,
        tokenizer_min_length: int = None,
        tokenizer_max_length: int = None,
        # Sentence Detector Params
        sentence_model_architecture: str = None,
        sentence_explode_sentences: bool = None,
        sentence_custom_bounds: list = None,
        sentence_use_custom_bounds_only: bool = None,
        sentence_split_length: int = None,
        sentence_min_length: int = None,
        sentence_max_length: int = None,
        sentence_impossible_penultimates: list = None,
        # Haystack Params
        progress_bar: bool = True,
    ):
        """
        :param split_mode: The split mode to determine how text should be segmented. Default: 'recursive'.
            It should be one of the following values:
                - "char": Split text based on individual characters.
                - "token": Split text based on tokens.
                - "sentence": Split text based on sentences.
                - "recursive": Split text recursively using a specific algorithm.
                - "regex": Split text based on a regular expression pattern.
        :param chunk_size: Size of each chunk of text. This param is applicable only for "recursive" split_mode.
        :param chunk_overlap: Length of the overlap between text chunks , by default `0`. This param is applicable only for "recursive" split_mode.
        :param split_patterns: Patterns to separate the text by in decreasing priority , by default `["\n\n", "\n", " ", ""]`.
        :param explode_splits: Whether to explode split chunks to separate rows , by default `True`.
        :param keep_separators: Whether to keep the separators in the final result , by default `True`.
        :param patterns_are_regex: Whether to interpret the split patterns as regular expressions , by default `False`.
        :param trim_whitespace: Whether to trim whitespaces of extracted chunks , by default `True`.
        :param sentence_awareness: Whether to split the document by sentence awareness if possible. Default: False.
            If true, it can stop the split process before maxLength. This param is not applicable only for "regex" and "recursive" split_mode.
        :param max_length: The maximum length allowed for spitting. The mode in which the maximum length is specified:
            - "char": Maximum length is measured in characters. Default: 512
            - "token": Maximum length is measured in tokens. Default: 128
            - "sentence": Maximum length is measured in sentences. Default: 8
        :param custom_bounds_strategy: The custom bounds strategy for text splitting using regular expressions.
            This param is applicable only for "regex" split_mode. It should be one of the following values:
                - "none": No custom bounds are applied.
                - "prepend": Custom bounds are prepended to the split documents.
                - "append": Custom bounds are appended to the split documents.
                - Default: "prepend".
        :param case_sensitive: Whether to use case sensitive when matching regex, by default `False`.
        :param meta_data_fields: Metadata fields to add specified data in columns to the metadata of the split documents.
            You should set column names to read columns.
        :param enable_sentence_increment: Whether the sentence index should be incremented in the metadata of the annotator. Default: False.

        Tokenizer parameters
        --------------------
        :param tokenizer_target_pattern: Pattern to grab from text as token candidates, by default ``\\S+``
        :param tokenizer_prefix_pattern: Regex with groups and begins with ``\\A`` to match target prefix, by default ``\\A([^\\s\\w\\$\\.]*)``
        :param tokenizer_suffix_pattern: Regex with groups and ends with ``\\z`` to match target suffix, by default ``([^\\s\\w]?)([^\\s\\w]*)\\z``
        :param tokenizer_infix_patterns: Regex patterns that match tokens within a single target. groups identify different sub-tokens. multiple defaults
        :param tokenizer_exceptions: Words that won't be affected by tokenization rules
        :param tokenizer_exceptions_path: Path to file containing list of exceptions
        :param tokenizer_case_sensitive_exceptions: Whether to care for case sensitiveness in exceptions, by default True
        :param tokenizer_context_chars: Character list used to separate from token boundaries, by default ['.', ',', ';', ':', '!', '?', '*', '-', '(', ')', '"', "'"]
        :param tokenizer_split_pattern: Pattern to separate from the inside of tokens. Takes priority over splitChars.
        :param tokenizer_split_chars: Character list used to separate from the inside of tokens
        :param tokenizer_min_length: The minimum allowed length for each token, by default 0
        :param tokenizer_max_length: The maximum allowed length for each token, by default 99999

        SentenceDetectorDLModel parameters
        ----------------------------------
        :param sentence_model_architecture: Model architecture (CNN)
        :param sentence_explode_sentences: Whether to explode each sentence into a different row, for better parallelization. Defaults to false.
        :param sentence_custom_bounds: Characters used to explicitly mark sentence bounds, by default []
        :param sentence_use_custom_bounds_only: Only utilize custom bounds in sentence detection, by default False
        :param sentence_split_length: Length at which sentences will be forcibly split
        :param sentence_min_length: The minimum allowed length for each sentence, by default 0
        :param sentence_max_length: The maximum allowed length for each sentence, by default 99999
        :param sentence_impossible_penultimates: Impossible penultimates - list of strings which a sentence can't end with

        :param progress_bar: Whether to show a progress bar.
        """
        super().__init__()
        self.progress_bar = progress_bar

        self.pipe = get_doc_splitter_internal_pipe(
            split_mode=split_mode,
            chunk_size=chunk_size,
            chunk_overlap=chunk_overlap,
            split_patterns=split_patterns,
            explode_splits=explode_splits,
            keep_separators=keep_separators,
            patterns_are_regex=patterns_are_regex,
            trim_whitespace=trim_whitespace,
            sentence_awareness=sentence_awareness,
            max_length=max_length,
            custom_bounds_strategy=custom_bounds_strategy,
            case_sensitive=case_sensitive,
            meta_data_fields=meta_data_fields,
            enable_sentence_increment=enable_sentence_increment,
            tokenizer_target_pattern=tokenizer_target_pattern,
            tokenizer_prefix_pattern=tokenizer_prefix_pattern,
            tokenizer_suffix_pattern=tokenizer_suffix_pattern,
            tokenizer_infix_patterns=tokenizer_infix_patterns,
            tokenizer_exceptions=tokenizer_exceptions,
            tokenizer_exceptions_path=tokenizer_exceptions_path,
            tokenizer_case_sensitive_exceptions=tokenizer_case_sensitive_exceptions,
            tokenizer_context_chars=tokenizer_context_chars,
            tokenizer_split_pattern=tokenizer_split_pattern,
            tokenizer_split_chars=tokenizer_split_chars,
            tokenizer_min_length=tokenizer_min_length,
            tokenizer_max_length=tokenizer_max_length,
            sentence_model_architecture=sentence_model_architecture,
            sentence_explode_sentences=sentence_explode_sentences,
            sentence_custom_bounds=sentence_custom_bounds,
            sentence_use_custom_bounds_only=sentence_use_custom_bounds_only,
            sentence_split_length=sentence_split_length,
            sentence_min_length=sentence_min_length,
            sentence_max_length=sentence_max_length,
            sentence_impossible_penultimates=sentence_impossible_penultimates,
        )


def inject():
    # inject the emd encoder into haystack
    from haystack.nodes.retriever import _embedding_encoder

    _embedding_encoder._EMBEDDING_ENCODERS[
        "johnsnowlabs"
    ] = _JohnsnowlabsEmbeddingEncoder
    # inject the retriever into haystack


inject()

# if id_hash_keys is None:
#     id_hash_keys = self.id_hash_keys
#
# if isinstance(document, dict):
#     document["id_hash_keys"] = id_hash_keys
#     document = Document.from_dict(document)
#
# # Mainly needed for type checking
# if not isinstance(document, Document):
#     raise HaystackError(
#         "Document must not be of type 'dict' but of type 'Document'."
#     )
#
# if type(document.content) is not str:
#     return document
#
# text = document.content
# # TODO APPLY PIPE
# self.pipe.annotate(text)
# texts = [split for split in self.pipe.annotate(text)["splits"]]
# if text != document.content:
#     document = deepcopy(document)
#     document.content = text
# return document



import os
import shutil
import sys
from pathlib import Path
from typing import Optional, List

from colorama import Fore

from johnsnowlabs import settings
from johnsnowlabs.auto_install.offline_install import get_py4j_dependency_urls
from johnsnowlabs.py_models import jsl_secrets
from johnsnowlabs.py_models.install_info import (
    JvmInstallInfo,
    PyInstallInfo,
    RootInfo,
    InstallSuite,
    LocalPy4JLib,
    InstallFolder,
)
from johnsnowlabs.py_models.jsl_secrets import JslSecrets
from johnsnowlabs.py_models.url_dependency import UrlDependency
from johnsnowlabs.utils.enums import (
    JvmHardwareTarget,
    PyInstallTypes,
    ProductName,
    ProductLogo,
)


def jsl_home_exist():
    return os.path.exists(settings.root_info_file)


def is_jsl_home_outdated():
    if jsl_home_exist:
        return (
            not RootInfo.get_from_jsl_home().version.as_str()
            == settings.raw_version_jsl_lib
        )
    else:
        raise Exception(f"JSL-Home does not exist! Cannot check if outdated")


def download_deps_and_create_info(
    deps: List[UrlDependency],
    lib_dir,
    info_file_path,
    overwrite=False,
):
    """Download a list of deps to given lib_dir folder and creates info_file at info_file_path."""
    info, old_info = {}, None
    if os.path.exists(info_file_path):
        #  keep old infos, we assume they are up-to-date and compatible
        if os.path.join("java_installs","info.json") in info_file_path:
            old_info = InstallFolder.java_folder_from_home()
        elif os.path.join("py_installs","info.json") in info_file_path:
            old_info = InstallFolder.py_folder_from_home()

    for p in deps:
        # print_prefix = Software.for_name(p.product_name).logo
        print_prefix = ProductLogo.from_name(p.product_name.name).value
        if p.dependency_type in JvmHardwareTarget:
            print_prefix = f"{ProductLogo.java.value}+{print_prefix} Java Library"
            constructor = JvmInstallInfo
        elif p.dependency_type in PyInstallTypes:
            print_prefix = f"{ProductLogo.python.value}+{print_prefix} Python Library"
            constructor = PyInstallInfo
        else:
            raise ValueError(f"Unknown Install type {p.dependency_type}")
        if not os.path.exists(f"{lib_dir}/{p.file_name}") or overwrite:
            try:
                p.download_url(lib_dir, name_print_prefix=print_prefix)
            except ValueError as _:
                # sys.tracebacklimit = 0
                err_msg = f"""🚨 Cannot install {ProductLogo.from_name(p.product_name.name).value}{p.product_name.value} because provided license file secret is outdated or it is invalid.
How to Fix this: 
Option1: Run {Fore.LIGHTGREEN_EX}nlp.install(force_browser=True){Fore.RESET} to get a Browser Window pop-up where you can refresh your license data
Option2: Run {Fore.LIGHTGREEN_EX}nlp.install(json_license_path="path/to/fresh_credentials.json"){Fore.RESET} after downloading a fresh license from https://my.johnsnowlabs.com/subscriptions   
Option3: Run {Fore.LIGHTGREEN_EX}nlp.install(refresh_install=True,force_browser=True){Fore.RESET} to refresh everything 
Option4: Set {Fore.LIGHTGREEN_EX}nlp.settings.enforce_versions=False{Fore.RESET} and run{Fore.LIGHTGREEN_EX} nlp.install(refresh_install=True,force_browser=True){Fore.RESET} to disable this protection mechanism and try to install anyways. Not Recommended, can yield unforeseen consequences 
"""

                print(err_msg)
                raise Exception(err_msg)

        info[p.file_name] = constructor(
            file_name=p.file_name,
            product=p.product_name,
            compatible_spark_version=p.spark_version.value.as_str(),
            install_type=p.dependency_type.value,
            product_version=p.dependency_version.as_str(),
        )
        info[p.file_name].compatible_spark_version = p.spark_version.value.as_str()
        info[p.file_name].product_version = p.dependency_version.as_str()

    if info:
        info = InstallFolder(**{"infos": info})
        if old_info:
            info.infos.update(old_info.infos)
        with open(info_file_path, "w") as f:
            for k, v in info.infos.items():
                v.product_version = str(v.product_version)
                v.compatible_spark_version = str(v.compatible_spark_version)

            f.write(info.model_dump_json())

def setup_jsl_home(
    secrets: Optional[JslSecrets] = None,
    jvm_install_type: JvmHardwareTarget = JvmHardwareTarget.cpu,
    py_install_type: PyInstallTypes = PyInstallTypes.wheel,
    only_jars: bool = False,
    spark_version=None,
    overwrite=False,
    log=True,
    refresh_install=False,
    visual=False,
    nlp=True,
    spark_nlp=True,
) -> None:
    """Folder structure :
    Creates Folder for JSL home and downloads all Py4J wheels/Jars
    for which we need to take PySpark Compatibility into account as well as JVM Hardware target
    ~.johnsnowlabs/
       ├─ licenses/
       │  ├─ info.json
       │  ├─ license1.json
       │  ├─ license2.json
       ├─ java_installs/
       │  ├─ info.json
       │  ├─ app1.jar
       │  ├─ app2.jar
       ├─ py_installs/
       │  ├─ info.json
       │  ├─ app1.tar.gz
       │  ├─ app2.tar.gz
       ├─ info.json
    """

    # Create all Paths
    Path(settings.license_dir).mkdir(parents=True, exist_ok=True)
    Path(settings.java_dir).mkdir(parents=True, exist_ok=True)
    Path(settings.py_dir).mkdir(parents=True, exist_ok=True)
    force_update = False
    suite = None

    if jsl_home_exist():
        if secrets:
            # Don't log because we will ignore the license from localhost, since one is provided
            jsl_secrets.already_logged = True
        suite = get_install_suite_from_jsl_home(
            create_jsl_home_if_missing=False,
            recursive_call=True,
            log=False,
            browser_login=False,
            jvm_hardware_target=jvm_install_type,
        )
        if secrets:
            # We overwrite secrets if user provided any
            suite.secrets = secrets

    if jsl_home_exist() and is_jsl_home_outdated() and log:
        print(f"🤓 Looks like {settings.root_dir} is outdated, updating it")
    if not jsl_home_exist() or is_jsl_home_outdated() or refresh_install:
        print(
            f"👷 Setting up  John Snow Labs home in {settings.root_dir}, this might take a few minutes."
        )
        # Delete everything except license data and re-create folder
        shutil.rmtree(settings.java_dir)
        shutil.rmtree(settings.py_dir)
        Path(settings.java_dir).mkdir(parents=True, exist_ok=True)
        Path(settings.py_dir).mkdir(parents=True, exist_ok=True)
        force_update = True

    # Get Urls for P4J based libs
    if force_update or suite and suite.get_missing_products(nlp, visual, spark_nlp):
        java_deps, py_deps = get_py4j_dependency_urls(
            secrets=secrets,
            spark_version=spark_version,
            visual=visual,
            nlp=nlp,
            spark_nlp=spark_nlp,
            jvm_install_type=jvm_install_type,
            py_install_type=py_install_type,
        )

        # store deps to jsl home with info.json files
        if not only_jars:
            download_deps_and_create_info(
                py_deps, settings.py_dir, settings.py_info_file, overwrite
            )
        download_deps_and_create_info(
            java_deps, settings.java_dir, settings.java_info_file, overwrite
        )

        root_info = RootInfo(version=settings.raw_version_jsl_lib, run_from=sys.executable)
        root_info.version = root_info.version.as_str()
        with open(settings.root_info_file, "w") as f:
            f.write(root_info.model_dump_json())
        print(f"🙆 JSL Home setup in {settings.root_dir}")

        return
    if log:
        print(f"👌 JSL-Home is up to date! ")


def get_install_suite_from_jsl_home(
    create_jsl_home_if_missing: bool = True,
    jvm_hardware_target: JvmHardwareTarget = JvmHardwareTarget.cpu,
    nlp: bool = True,
    visual: bool = False,
    spark_nlp: bool = True,
    only_jars: bool = False,
    recursive_call=False,
    # Secret Flow Params
    browser_login: bool = True,
    force_browser: bool = False,
    access_token: Optional[str] = None,
    local_license_number: int = 0,
    remote_license_number: int = 0,
    secrets_file: Optional[str] = None,
    hc_license: Optional[str] = None,
    hc_secret: Optional[str] = None,
    ocr_secret: Optional[str] = None,
    ocr_license: Optional[str] = None,
    aws_access_key: Optional[str] = None,
    aws_key_id: Optional[str] = None,
    fin_license: Optional[str] = None,
    leg_license: Optional[str] = None,
    store_in_jsl_home: bool = True,
    log: bool = True,
) -> InstallSuite:
    """Read all info files from JSL home if exists. If not exists, sets up JSL home"""
    if not jsl_home_exist() and not create_jsl_home_if_missing:
        return InstallSuite.empty()
    license_data: JslSecrets = JslSecrets.build_or_try_find_secrets(
        browser_login=browser_login,
        force_browser=force_browser,
        access_token=access_token,
        local_license_number=local_license_number,
        remote_license_number=remote_license_number,
        secrets_file=secrets_file,
        hc_license=hc_license,
        hc_secret=hc_secret,
        ocr_secret=ocr_secret,
        ocr_license=ocr_license,
        aws_access_key=aws_access_key,
        aws_key_id=aws_key_id,
        return_empty_secrets_if_none_found=True,
        fin_license=fin_license,
        leg_license=leg_license,
        store_in_jsl_home=store_in_jsl_home,
    )

    if create_jsl_home_if_missing:
        if not jsl_home_exist():
            # Nothing setup yet, download at least spark nlp jars
            print(f"🤓 Looks like {settings.root_dir} is missing, creating it")
            setup_jsl_home(only_jars=only_jars, log=False)

        if jsl_home_exist() and is_jsl_home_outdated():
            # Nothing setup yet, download at least spark nlp jars
            setup_jsl_home(only_jars=only_jars, log=False)

    java_folder, py_folder = None, None

    if os.path.exists(settings.java_info_file):
        java_folder = InstallFolder.java_folder_from_home()
    if os.path.exists(settings.py_info_file):
        py_folder = InstallFolder.py_folder_from_home()

    info = RootInfo.get_from_jsl_home()
    # Read all dependencies from local ~/.johnsnowlabs folder
    suite = InstallSuite(
        nlp=LocalPy4JLib(
            java_lib=java_folder.get_product_entry(ProductName.nlp, jvm_hardware_target)
            if java_folder
            else None,
            py_lib=py_folder.get_product_entry(ProductName.nlp) if py_folder else None,
        ),
        hc=LocalPy4JLib(
            java_lib=java_folder.get_product_entry(ProductName.hc)
            if java_folder
            else None,
            py_lib=py_folder.get_product_entry(ProductName.hc) if py_folder else None,
        ),
        ocr=LocalPy4JLib(
            java_lib=java_folder.get_product_entry(ProductName.ocr)
            if java_folder
            else None,
            py_lib=py_folder.get_product_entry(ProductName.ocr) if py_folder else None,
        ),
        secrets=license_data,
        info=info,
    )

    missing = suite.get_missing_products(nlp, visual, spark_nlp)
    if missing and recursive_call and log:
        print(f"🚨 Looks like some of the missing jars could not be fetched...")
        suite.log_missing_jars(visual, nlp, spark_nlp)

    if missing and not recursive_call:
        print(f"🤓 Looks like you are missing some jars, trying fetching them ...")
        setup_jsl_home(
            license_data,
            jvm_install_type=jvm_hardware_target,
            only_jars=only_jars,
            log=False,
            nlp=nlp,
            visual=visual,
            spark_nlp=spark_nlp,
        )
        # After re-setting up jsl_home, call this method again
        return get_install_suite_from_jsl_home(
            jvm_hardware_target=jvm_hardware_target,
            nlp=nlp,
            visual=visual,
            spark_nlp=spark_nlp,
            only_jars=only_jars,
            recursive_call=True,
            browser_login=browser_login,
            access_token=access_token,
            local_license_number=local_license_number,
            remote_license_number=remote_license_number,
            secrets_file=secrets_file,
            hc_license=hc_license,
            hc_secret=hc_secret,
            ocr_secret=ocr_secret,
            ocr_license=ocr_license,
            aws_access_key=aws_access_key,
            aws_key_id=aws_key_id,
            fin_license=fin_license,
            leg_license=leg_license,
        )
    return suite



import os
from os.path import expanduser

from johnsnowlabs.utils.env_utils import (
    env_required_license,
    is_running_in_databricks_runtime,
    is_running_in_emr,
    set_py4j_logger_to_error_on_databricks,
)

# These versions are used for auto-installs and version  checks

raw_version_jsl_lib = "5.5.2rc1"

raw_version_nlp = "5.5.1"

raw_version_nlu = "5.4.1"


raw_version_pyspark = "3.4.0"
raw_version_nlp_display = "5.0"

raw_version_medical = "5.5.0"
raw_version_secret_medical = "5.5.0"

raw_version_secret_ocr = "5.4.2"
raw_version_ocr = "5.4.2"

raw_version_pydantic = "2"

pypi_page = "https://pypi.org/project/johnsnowlabs"
json_indent = 4
enforce_secret_on_version = False
enforce_versions = True

# Environment
on_databricks = is_running_in_databricks_runtime()
on_emr = is_running_in_emr()
license_required = env_required_license()

# Set root path, from which all other paths will be relative

if on_databricks:
    root_dir = f"/dbfs/johnsnowlabs"
    set_py4j_logger_to_error_on_databricks()
elif on_emr:
    root_dir = f"/lib/.johnsnowlabs/johnsnowlabs"
else:
    home = expanduser("~")
    if home[-1] == "/":
        # on Docker and other setups, ~ may resolve to `/`, so we need special handling here
        root_dir = f'{expanduser("~")}.johnsnowlabs'
    else:
        root_dir = f'{expanduser("~")}/.johnsnowlabs'

## Directories
license_dir = os.path.join(root_dir, "licenses")
java_dir = os.path.join(root_dir, "java_installs")
py_dir = os.path.join(root_dir, "py_installs")
tmp = os.path.join(root_dir, "tmp")

# Info Files
root_info_file = os.path.join(root_dir, "info.json")
java_info_file = os.path.join(java_dir, "info.json")
py_info_file = os.path.join(py_dir, "info.json")
creds_info_file = os.path.join(license_dir, "info.json")

# databricks paths
dbfs_home_dir = "dbfs:/johnsnowlabs"
dbfs_java_dir = f"{dbfs_home_dir}/java_installs"
dbfs_py_dir = f"{dbfs_home_dir}/py_installs"
db_py_jobs_dir = f"{dbfs_home_dir}/py_jobs"
db_py_notebook_dir = f"{dbfs_home_dir}/py_notebook_jobs"
db_jar_jobs_dir = f"{dbfs_home_dir}/jar_jobs"

db_cluster_name = "John-Snow-Labs-Databricks-Auto-Cluster🚀"
db_driver_node_type = "i3.xlarge"
db_node_type_id = "i3.xlarge"
db_spark_version = "10.5.x-scala2.12"

# db_spark_version = "13.2.x-scala2.12"

db_job_name = "John-Snow-Labs-Job {job} 🚀"
db_run_name = "John-Snow-Labs-Run 🚀"
db_driver_node_type = "i3.xlarge"

# emr settings
emr_cluster_name = "John-Snow-Labs-EMR-Auto-Cluster🚀"
emr_release_label = "emr-6.5.0"
emr_instance_type = "m4.4xlarge"
emr_instance_count = 3
emr_applications = ["Hadoop", "Spark", "Livy", "JupyterEnterpriseGateway"]
emr_default_instance_profile = "EMR_EC2_DefaultRole"
emr_default_service_role = "EMR_DefaultRole"
emr_volume_size = 100

# Local Spark mode
spark_session_name = "John-Snow-Labs-Spark-Session 🚀"


### Docker Settings

docker_image_name = "johnsnowlabs_image"
docker_container_name = "johnsnowlabs_container"



from nbformat import NotebookNode


class Nodes:
    # Re-Usable Nodes to Inject
    sparksession = NotebookNode(
        cell_type="code",
        source="from johnsnowlabs import *;spark = nlp.start()",
        execution_count=0,
        metadata={"collapsed": True, "pycharm": {"name": "#%%\n"}},
        outputs=[],
    )
    sparksession_with_cache_folder = NotebookNode(
        cell_type="code",
        source='from johnsnowlabs import *;spark = nlp.start(cache_folder="/home/ckl/dump/cache_pretrained")',
        execution_count=0,
        metadata={"collapsed": True, "pycharm": {"name": "#%%\n"}},
        outputs=[],
    )

    imports = NotebookNode(
        cell_type="code",
        source="from sparknlp.base import LightPipeline\n"
        "from pyspark.ml import Pipeline",
        execution_count=0,
        metadata={"collapsed": True, "pycharm": {"name": "#%%\n"}},
        outputs=[],
    )

    dummy = NotebookNode(
        cell_type="code",
        source='print("HELLO DUMMY WORLD")',
        execution_count=0,
        metadata={"collapsed": True, "pycharm": {"name": "#%%\n"}},
        outputs=[],
    )



import os
from random import randrange
from typing import Optional

from johnsnowlabs.utils.boto_utils import BotoException
from johnsnowlabs.utils.file_utils import path_tail
from johnsnowlabs.utils.s3_utils import (
    check_if_file_exists_in_s3,
    create_bucket,
    upload_content,
    upload_file_to_s3,
)


def create_emr_bucket(boto_session: "boto3.Session", bucket=None):
    """Create a bucket for EMR cluster logs
    :param boto_session: Boto3 session
    :param bucket: Bucket name
    """
    import botocore

    try:
        sts_client = boto_session.client("sts")
        account_id = sts_client.get_caller_identity()["Account"]
        region = sts_client.meta.region_name
        if not bucket:
            bucket = f"johnsnowlabs-emr-{account_id}-{region}"

        return create_bucket(boto_session=boto_session, bucket=bucket)

    except botocore.exceptions.ClientError as e:
        if e.response["Error"]["Code"] == "BucketAlreadyOwnedByYou":
            return bucket
        raise BotoException(
            code=e.response["Error"]["Code"], message=e.response["Error"]["Message"]
        )


def run_in_emr(
    py_script_path: str,
    cluster_id: str,
    boto_session: Optional["boto3.Session"] = None,
    script_bucket: Optional[str] = None,
    bucket_folder_path: Optional[str] = None,
    run_name: Optional[str] = None,
    execution_role_arn: Optional[str] = None,
) -> str:
    """Run a python script in EMR cluster
    :param boto_session: Boto3 session. Refer to https://boto3.amazonaws.com/v1/documentation/api/latest/reference/core/session.html
    :param py_script_path: Script content or path to script file in local file system or S3
    :param cluster_id: EMR cluster id
    :param script_bucket: S3 bucket where the script will be uploaded
    :param bucket_folder_path: S3 bucket folder where the script will be uploaded
    :param run_name: Name of the step
    :param execution_role_arn: IAM role to use for the step
    :return: Step id
    """
    import boto3
    import botocore

    if not boto_session:
        boto_session = boto3.Session()
    if check_if_file_exists_in_s3(py_script_path):
        s3_path = py_script_path
    else:
        # Making sure bucket exists

        script_bucket = create_emr_bucket(
            boto_session=boto_session, bucket=script_bucket
        )

        script_s3_path = temporary_script_name = f"{randrange(1333777)}tmp.py"
        if bucket_folder_path is not None:
            script_s3_path = os.path.join(bucket_folder_path, temporary_script_name)
        local_script = os.path.expanduser(py_script_path)
        if os.path.exists(local_script):
            s3_path = upload_file_to_s3(
                boto_session=boto_session,
                file_path=local_script,
                bucket=script_bucket,
                file_name=script_s3_path,
            )
        else:
            s3_path = upload_content(
                boto_session=boto_session,
                content=py_script_path,
                bucket=script_bucket,
                file_name=script_s3_path,
            )

    payload = {
        "JobFlowId": cluster_id,
        "Steps": [
            {
                "Name": run_name or path_tail(s3_path),
                "ActionOnFailure": "CONTINUE",
                "HadoopJarStep": {
                    "Jar": "command-runner.jar",
                    "Args": [
                        "spark-submit",
                        "--master",
                        "yarn",
                        s3_path,
                    ],
                },
            }
        ],
    }
    if execution_role_arn:
        payload["ExecutionRoleArn"] = execution_role_arn

    try:
        try:
            emr_client = boto_session.client("emr")
            response = emr_client.add_job_flow_steps(**payload)
            return response["StepIds"][0]
        except botocore.exceptions.ClientError as e:
            raise BotoException(
                code=e.response["Error"]["Code"], message=e.response["Error"]["Message"]
            )
    except Exception as e:
        print("❌ Failed running script in EMR cluster. Error: ", e)



import json
import time

from johnsnowlabs import nlp
from johnsnowlabs.auto_install.docker.work_utils import check_local_endpoint_health, _destroy_container, _destroy_image
from johnsnowlabs.utils.py_process import run_cmd_and_check_succ


def is_snowflake_installed():
    try:
        import snowflake.connector
    except:
        ImportError('Run `pip install snowflake-connector-python` to use Snowflake utilities! ')
    return True


def get_client(user, password, account, warehouse, database, schema, role):
    is_snowflake_installed()
    import snowflake.connector
    conn = snowflake.connector.connect(
        user=user,
        password=password,
        account=account,
        warehouse=warehouse,
        database=database,
        schema=schema,
        role=role,
    )
    return conn


def query_udf(client, udf_name, data):
    cmd_query_udf = """SELECT {udf_name}('{data}')"""
    cur = client.cursor()
    cur.execute(cmd_query_udf.format(udf_name=udf_name, data=data))
    for row in cur:
        data = json.loads(row[0])
        print(data)
    cur.close()
    return data


def get_service_create_cmd(service_name, compute_pool_name, image_path, role, database, warehouse, schema):
    return f"""
    USE ROLE {role};
    USE DATABASE {database};
    USE WAREHOUSE {warehouse};
    USE SCHEMA {schema};
    
    CREATE SERVICE {service_name}
      IN COMPUTE POOL {compute_pool_name}
      FROM SPECIFICATION $$
        spec:
          containers:
          - name: jsl-container
            image: {image_path}
            readinessProbe:
              port: 80
              path: /ping
          endpoints:
          - name: invoke
            port: 80
            public: true
            $$
       MIN_INSTANCES=1
       MAX_INSTANCES=1;
    """


def build_snowflake_image(nlu_ref, image_name, license_path):
    # check_build_serve_query
    nlp.build_image(
        nlu_ref,
        image_name,
        rebuild=True,
        use_cache=True,
        json_license_path=license_path,
    )


def test_snowflake_image_local(image_name, container_name, port):
    # Serve container, destroy if already running. After test destroy local container
    nlp.serve_container(
        destroy_container=True,
        image_name=image_name,
        container_name=container_name,
        host_port=port,
    )
    # todo expo backoff for big models N times maybe
    time.sleep(30)
    check_local_endpoint_health(port)
    _destroy_container(container_name)


def push_snowflake_image(remote_repo, image_name):
    cmd = f'docker push {remote_repo}/{image_name}:latest'
    return run_cmd_and_check_succ(
        [cmd], shell=True, raise_on_fail=True, use_code=True, log=True
    )


def wait_until_service_created(client, service, timeout=60*20):
    cur = client.cursor()
    start_time = time.time()  # Start time tracking

    while True:
        elapsed_time = time.time() - start_time  # Calculate elapsed time
        if elapsed_time > timeout:
            raise TimeoutError(f"Timeout reached: Service {service} did not reach RUNNING state within {timeout} seconds.")

        res = cur.execute(f'DESCRIBE SERVICE {service};').fetchone()
        state = res[1]

        if state == 'RUNNING':
            return True
        elif state == 'PENDING':
            pass
        elif state in ['FAILED', 'DONE', 'SUSPENDING', 'SUSPENDED', 'DELETING', 'DELETED', 'INTERNAL_ERROR']:
            raise Exception(f"State of service {service} is {state}; cannot continue.")

        print(f"State of {service} is {state}. Waiting for service creation...")
        time.sleep(60)



def create_service(client, service_name, compute_pool_name, image_path, role, database, warehouse, schema):
    cur = client.cursor()
    cmd = get_service_create_cmd(service_name, compute_pool_name, image_path, role, database, warehouse, schema)
    try:
        cur.execute(cmd, num_statements=cmd.count(';'))
    except Exception as e:
        if 'already exist' in e.raw_msg:
            print(f'A resource already exists, see error message for more details')
            print(e.raw_msg)
        raise Exception("Snowflake Service creation failed see error message for more details")
    for row in cur:
        print(row)
    cur.close()
    print('service created')


def create_udf(client, service_name, udf_name, role, database, warehouse, schema):
    cur = client.cursor()
    cmd = create_udf_cmd(service_name, udf_name, role, database, warehouse, schema)
    cur.execute(cmd, num_statements=cmd.count(';'))
    for row in cur:
        print(row)
    cur.close()


def create_udf_cmd(service_name, udf_name, role, database, warehouse, schema):
    return f'''
USE ROLE {role};
USE DATABASE {database};
USE WAREHOUSE {warehouse};
USE SCHEMA {schema};

CREATE FUNCTION {udf_name} (InputText varchar, OutputLevel VARCHAR DEFAULT 'document')
  RETURNS object
  SERVICE={service_name}
  ENDPOINT=invoke
  AS '/invoke';
    '''


def test_udf(client, udf_name):
    # this will test service under the hood
    return query_udf(client, udf_name, 'Hello this is my data ')


def tag_image(image_name, remote_repo):
    cmd = f'docker tag {image_name}:latest {remote_repo}/{image_name}:latest'
    return run_cmd_and_check_succ(
        [cmd], shell=True, raise_on_fail=True, use_code=True, log=True
    )


def build_test_and_push_image(nlu_ref, license_path, image_name, local_test_container_name, local_test_port,
                              remote_repo, user, password, destroy_image=False):
    login_cmd = f"echo {password} | docker login {remote_repo} -u {user} --password-stdin"

    # 1. build image,
    build_snowflake_image(nlu_ref, image_name, license_path)

    # 2. test it locally
    test_snowflake_image_local(image_name, local_test_container_name, local_test_port)

    # 3. tag it
    tag_image(image_name, remote_repo)

    # 4. login to docker repo,
    print(f'Logging into repo {remote_repo} with user {user}')
    run_cmd_and_check_succ(login_cmd, text=True, shell=True, use_code=True, log=False)

    # 5. push image
    push_snowflake_image(remote_repo, image_name)

    # 6. destroy the local image
    if destroy_image:
        _destroy_image(image_name)


def get_service_logs(snowflake_user, snowflake_password, snowflake_account, warehouse_name, database_name,
                     schema_name, role_name, service_name):
    client = get_client(snowflake_user, snowflake_password, snowflake_account, warehouse_name, database_name,
                        schema_name, role_name)
    cur = client.cursor()
    r = cur.execute(f"SELECT SYSTEM$GET_SERVICE_STATUS('{service_name}');").fetchall()
    print(r)
    import json
    status = json.loads(r[0][0])[0]['status']
    if status == "PENDING":
        return True
    else:
        pass
    return r


def snowflake_common_setup(snowflake_user, snowflake_account, snowflake_password,
                           role_name='test_role',
                           schema_name='data_schema',
                           repo_name='tutorial_repository',
                           stage_name='tutorial_stage',
                           db_name='tutorial_db',
                           warehouse_name='tutorial_warehouse',
                           compute_pool_name='tutorial_compute_pool',
                           ):
    """do commmon setup for Snowflake Container Services.
    Creates Warehouse, Database, Schema, Compute-Pool, Repository and Role to use for Johnsnowlabs Based container services.

    """

    # todo params for warehouse_name size and compute-pool
    base_cmd = f"""
USE ROLE ACCOUNTADMIN;
USE WAREHOUSE {warehouse_name};
CREATE ROLE IF NOT EXISTS {role_name};

CREATE DATABASE IF NOT EXISTS {db_name};
GRANT OWNERSHIP ON DATABASE {db_name} TO ROLE {role_name} COPY CURRENT GRANTS;

CREATE OR REPLACE WAREHOUSE {warehouse_name} WITH WAREHOUSE_SIZE='X-SMALL';
GRANT USAGE ON WAREHOUSE {warehouse_name} TO ROLE {role_name};


GRANT BIND SERVICE ENDPOINT ON ACCOUNT TO ROLE {role_name};

CREATE COMPUTE POOL IF NOT EXISTS {compute_pool_name} 
  MIN_NODES = 1
  MAX_NODES = 1
  INSTANCE_FAMILY = CPU_X64_XS;
  
GRANT USAGE, MONITOR ON COMPUTE POOL {compute_pool_name} TO ROLE {role_name};

GRANT ROLE {role_name} TO USER {snowflake_user};
    
    """

    is_snowflake_installed()
    import snowflake.connector
    c = snowflake.connector.connect(user=snowflake_user, password=snowflake_password, account=snowflake_account, role='ACCOUNTADMIN')
    cur = c.cursor()

    r = cur.execute(base_cmd, num_statements=base_cmd.count(';'))
    succ = r.fetchall()[0][0] == 'Statement executed successfully.'
    if succ:
        print(f'Created Role {role_name} and access granted to {snowflake_user}')
        print(f'Created Database {db_name}')
        print(f'Created Warehouse {warehouse_name}')
        print(f'Created Compute Pool {compute_pool_name}')

    create_db_objects_cmd = f"""
USE ROLE {role_name};
USE DATABASE {db_name};
USE WAREHOUSE {warehouse_name};

CREATE SCHEMA IF NOT EXISTS {schema_name};
USE SCHEMA  {schema_name};
CREATE IMAGE REPOSITORY IF NOT EXISTS {repo_name};
CREATE STAGE IF NOT EXISTS {stage_name} DIRECTORY = ( ENABLE = true );
    """

    r = cur.execute(create_db_objects_cmd, num_statements=create_db_objects_cmd.count(';'))

    succ = r.fetchall()[0][0] == 'Statement executed successfully.'
    if succ:
        print(f'Created Schema {schema_name}')
        print(f'Created Repository {repo_name}')
        print(f'Created Stage {stage_name}')


    else:
        print('Failure creating Schema, Repository and Stage!')

    verify_prefix = f'''
USE ROLE {role_name};
USE DATABASE {db_name};
USE WAREHOUSE {warehouse_name};
    '''

    def verify_image_repo(verify_prefix):
        cmd = f'SHOW IMAGE REPOSITORIES;'
        response = cur.execute(cmd, num_statements=cmd.count(';')).fetchall()
        print(response)
        for r in response:
            response_repo = r[1].lower()
            response_db = r[2].lower()
            response_schema = r[3].lower()
            response_repo_url = r[4].lower()
            response_role = r[5].lower()
            if response_repo == repo_name.lower() and response_db == db_name.lower() and response_schema == schema_name.lower() \
                    and response_role == role_name.lower():
                return response_repo_url

    repo_url = verify_image_repo(verify_prefix)
    print(f'Created Snowflake Container Repository {repo_url}')
    return role_name, db_name, warehouse_name, schema_name, compute_pool_name, repo_url


def deploy_as_snowflake_udf(nlu_ref,
                            repo_url,
                            role_name,
                            database_name,
                            warehouse_name,
                            schema_name,
                            compute_pool_name,
                            snowflake_user,
                            snowflake_password,
                            snowflake_account,
                            license_path=None,
                            udf_name=None,
                            service_name=None,
                            service_creation_timeout = 60*20,
                            destroy_image=False,
                            ):
    client = get_client(snowflake_user, snowflake_password, snowflake_account, warehouse_name, database_name,
                        schema_name, role_name)

    # Local container setup
    clean_nlu_ref = nlu_ref.replace('.', '-').replace('_', '-')

    port = 6645
    container_name = f"{clean_nlu_ref}_container"
    image_name = f"{clean_nlu_ref}-img"

    remote_image = f'{repo_url}/{image_name}:latest'

    if not service_name:
        service_name = f'{clean_nlu_ref}_service'.replace('-', '_')
    if not udf_name:
        udf_name = f'{clean_nlu_ref}_udf'.replace('-', '_')

    # r = get_service_logs(snowflake_user, snowflake_password, snowflake_account, warehouse_name, database_name,
    #              schema_name, role_name, service_name)
    # print(r)
    # 2.  Local Docker Setup, Tests and Push to Snowflake
    build_test_and_push_image(nlu_ref, license_path, image_name, container_name, port, repo_url, snowflake_user, snowflake_password, destroy_image)

    # 3. Snowflake: Create service, create udf and test udf
    print(f'Starting Snowflake Procedure')
    create_service(client, service_name, compute_pool_name, remote_image, role_name, database_name, warehouse_name,
                   schema_name)
    print(f'Created Service {service_name}')
    wait_until_service_created(client, service_name, service_creation_timeout)
    # time.sleep(1 * 60)  # wait ~ n seconds for container sto spin up, expo backup..!
    create_udf(client, service_name, udf_name, role_name, database_name, warehouse_name, schema_name)
    print(f'Created UDF {udf_name}')

    print('testing UDF...')
    test_udf(client, udf_name)

    return udf_name





import os
import shutil
import sys
from pathlib import Path
from typing import Dict, Set

from johnsnowlabs import settings
from johnsnowlabs.abstract_base.software_product import AbstractSoftwareProduct
from johnsnowlabs.auto_install.softwares import Software
from johnsnowlabs.py_models.jsl_secrets import JslSecrets
from johnsnowlabs.utils.enums import ProductName, LatestCompatibleProductVersion
from johnsnowlabs.utils.env_utils import try_import
from johnsnowlabs.utils.pip_utils import get_pip_lib_version
from johnsnowlabs.utils.venv_utils import VenvWrapper


def check_and_install_dependencies(
    product: AbstractSoftwareProduct,
    secrets: JslSecrets,
    python_exec_path: str = sys.executable,
    install_optional: bool = False,
    install_licensed: bool = True,
    py_setup_dir: str = None,
    offline_zip_dir: str = None,
    include_dependencies: bool = True,
    visual: bool = False,
    spark_nlp: bool = True,
    enterpise_nlp: bool = True,
):
    """
    Iterates the dependency DAG in DFS order for input product and downloads installs all dependencies
    into python_exec_path with pip module.
    Usually we just iterate starting from jsl-full-suite Product, which depends on everything

    A product may have 3 types of dependencies which are defined as outgoing edges in the DAG:
        - hard_dependency: Must have this installed for core functionality like pyspark
        - licensed_dependency: Requires a license, install if accessible like OCR/NLP-Licensed
        - optional_dependency: Can be installed for extra features, like NLU, NLP-Display, etc..

    NOTE: There is currently no product that has licensed dependencies.
    INVARIANT :
        Either the product itself is licensed and then no dependencies requiring license
        Or product is free, but it has licensed optional dependencies

    Spark NLP must be re-installed last, because other libs may depend on it but might install a
    Spark NLP version which is not the one we have defined in johnsnowlabs.settings

    TODO (UX Improvement)
        To properly log what was installed, we must first iterate the DAG and get all missing notes
        Otherwise we might install multiple nodes because of sub-dependencies in one iteration
        And then we will not be able to properly track what was installed ( Everything works fine tho)
        This causes venv/cli based installs which will install jsl_lib not to print out all JSL-product logos
            Like Spark-NLP-Display, NLU, Pyspark or Spark-NLP

    # TODO generalize HOST class Databricks/Local/Venv etc.. with host.has_lib_installed() etc..

    :param offline_zip_dir:
    :param include_dependencies:
    :param py_dir:
    :param product:
    :param python_exec_path: A Python executable into which products should be installed.
        The pip module of this executable is used to install all libs
    :param py_setup_dir: If not None, a fresh Python Venv with target products will be setup in folder
        if using this parameter, the python_exec_path parameter will be ignored
        and the python_exec_path from the newly created venv is used to setup all libs
    :param secrets: to use for installing licensed libs
    :param install_optional: install optional dependencies if True, otherwise not
    :param install_licensed: install licensed products if license permits it if True, otherwise not
    sparknlp_to_latest: for some releases we might not want to go to the latest spark release
    """
    import site
    from importlib import reload

    reload(site)
    offline_py_dir = None
    license_dir = None
    java_dir = None
    if py_setup_dir:
        # Use VenvWrapper to setup new Env
        VenvWrapper.create_venv(venv_target_dir=py_setup_dir, log=False)
        python_exec_path = VenvWrapper.glob_py_exec_from_venv(py_setup_dir)
    if offline_zip_dir and not os.path.exists(offline_zip_dir):
        offline_py_dir = f"{offline_zip_dir}/py_installs"
        java_dir = f"{offline_zip_dir}/java_installs"
        license_dir = f"{offline_zip_dir}/licenses"
        Path(offline_py_dir).mkdir(parents=True, exist_ok=True)
        Path(java_dir).mkdir(parents=True, exist_ok=True)
        Path(license_dir).mkdir(parents=True, exist_ok=True)

    # Iteration Variables
    hard_nodes: Set[AbstractSoftwareProduct] = {product}
    licensed_nodes: Set[AbstractSoftwareProduct] = set([])
    optional_nodes: Set[AbstractSoftwareProduct] = set([])
    install_results: Dict[AbstractSoftwareProduct:bool] = {}

    # Boolean Checkers
    is_spark_nlp = lambda node: node.name == ProductName.nlp.value
    is_ocr = lambda node: node.name == ProductName.ocr.value
    is_healthcare = lambda node: node.name == ProductName.hc.value
    exist_install_result = lambda node: node in install_results
    licensed_nodes_left_to_install = lambda: licensed_nodes and install_licensed
    optional_nodes_left_to_install = lambda: optional_nodes and install_optional

    while (
        hard_nodes
        or licensed_nodes_left_to_install()
        or optional_nodes_left_to_install()
    ):
        # Core loop, check if any more vertex left to explore
        # In addition to popping whatever is not empty, we remember what we popped by returning a tuple
        v = (
            hard_nodes.pop()
            if hard_nodes
            else licensed_nodes.pop()
            if licensed_nodes_left_to_install()
            else optional_nodes.pop()
            if optional_nodes_left_to_install()
            else None
        )

        if not visual and is_ocr(v):
            continue
        if not enterpise_nlp and is_healthcare(v):
            continue
        if not spark_nlp and is_spark_nlp(v):
            continue

        v: AbstractSoftwareProduct
        # Collect all children of this vertex for next iteration
        hard_nodes = hard_nodes | v.hard_dependencies
        licensed_nodes = licensed_nodes | v.licensed_dependencies
        optional_nodes = optional_nodes | v.optional_dependencies

        # Check if we should install this vertex
        if not v.pypi_name:
            # Non Python Dependencies are not handled here yet
            continue
        if not v.licensed and not install_licensed:
            continue
        if (
            v.check_installed(
                python_exec_path=python_exec_path, download_folder=offline_py_dir
            )
            and v.check_installed_correct_version()
            and not offline_py_dir
        ):
            # It's already installed and has correct version
            continue
        elif is_spark_nlp(v) and not offline_py_dir:
            # We don't install spark nlp during iterating the DAG
            continue
        elif exist_install_result(v):
            # We could have failed in a previous iteration installing this vertex
            # Only if we don't have an entry for it, we attempt installing
            continue

        # Attempt installing the node and store the result
        install_results[v] = v.install(
            secrets=secrets,
            py_path=python_exec_path,
            download_folder=offline_py_dir,
            include_dependencies=include_dependencies,
            re_install=v.check_installed_correct_version(),
        )
    if offline_zip_dir:
        print("👷 Zipping installation files for offline install")
        finalize_zip_folder(offline_zip_dir, java_dir, offline_py_dir, license_dir)
        print("Done zipping")
        return
    else:
        # These checks are only for current env, not defined for offline
        if not get_pip_lib_version("spark-nlp", py_exec=python_exec_path).equals(
            LatestCompatibleProductVersion.spark_nlp.value
        ):
            # Re-install NLP incase some other library up/downgraded it while we installed it
            install_results[Software.spark_nlp] = Software.spark_nlp.install(
                re_install=True,
                version=LatestCompatibleProductVersion.spark_nlp.value,
                py_path=python_exec_path,
                download_folder=offline_py_dir,
                include_dependencies=include_dependencies,
            )
        if not get_pip_lib_version("pyspark", py_exec=python_exec_path).equals(
            LatestCompatibleProductVersion.pyspark.value
        ):
            # Re-install pyspark incase some other library up/downgraded it while we installed it
            install_results[Software.pyspark] = Software.pyspark.install(
                re_install=True,
                version=LatestCompatibleProductVersion.pyspark.value,
                py_path=python_exec_path,
                download_folder=offline_py_dir,
                include_dependencies=include_dependencies,
            )

    # Log The results of installation
    if Software.jsl_full in install_results:
        del install_results[Software.jsl_full]
    if len(install_results) > 0:
        print(f"Installed {len(install_results)} products:")
        for installed_software, result in install_results.items():
            if installed_software.check_installed(
                python_exec_path=python_exec_path, download_folder=offline_py_dir
            ):
                print(
                    f"{installed_software.logo} {installed_software.name}=={installed_software.get_installed_version(prefer_pip=True, fallback_import=False)}"
                    f" installed! ✅ {installed_software.slogan} "
                )
            else:
                print(
                    f"{installed_software.logo} {installed_software.name} not installed! ❌"
                )

        # Trigger Imports after install. so module is reloaded and no re-start is required
        from johnsnowlabs import medical, finance, legal, visual, nlp
        from importlib import reload

        # reload all module which m ight have changes
        to_reload = [medical, finance, legal, visual, nlp]
        if try_import("PIL"):
            import PIL

            to_reload.append(PIL)
        for mod in to_reload:
            reload(mod)

    else:
        print(f"👌 Everything is already installed, no changes made")


def finalize_zip_folder(offline_zip_dir, new_java_dir, py_dir, new_license_dir):
    # Copy files from local jsl home and create zip file
    copy_jars_from_jsl_home_to_offline_zip_dir(new_java_dir)
    copy_py_installs_from_jsl_home_to_offline_zip_dir(py_dir)
    copy_licenses_from_jsl_home_to_zip_dir(new_license_dir)
    zip_folder(offline_zip_dir)


def copy_py_installs_from_jsl_home_to_offline_zip_dir(new_py_dir):
    for f in os.listdir(settings.py_dir):
        if ".gz" in f or ".whl" in f:
            print(f"Adding {f} to zip")
            shutil.copy(f"{settings.py_dir}/{f}", new_py_dir)


def copy_jars_from_jsl_home_to_offline_zip_dir(new_java_dir):
    for f in os.listdir(settings.java_dir):
        if ".jar" in f:
            print(f"Adding {f} to zip")
            shutil.copy(f"{settings.java_dir}/{f}", new_java_dir)


def copy_licenses_from_jsl_home_to_zip_dir(new_license_dir):
    for f in os.listdir(settings.license_dir):
        if "info" not in f:
            shutil.copy(f"{settings.license_dir}/{f}", new_license_dir)


def zip_folder(offline_zip_dir):
    shutil.make_archive("john_snow_labs_suite", "zip", offline_zip_dir)
    shutil.move("john_snow_labs_suite.zip", offline_zip_dir)



from typing import Optional


class LibVersion:
    """Representation of a library version in format A.B.C
    where elements can digits or X which matches all others digits"""

    def __init__(
        self,
        major_or_canonical_str: str,
        minor: Optional[str] = None,
        patch: Optional[str] = None,
    ):
        self.major, self.minor, self.patch = None, None, None

        if "." in major_or_canonical_str:

            splits = major_or_canonical_str.lower().split(".")
            if len(splits) == 0:
                raise ValueError(
                    "When using canonical representation to construct a LibVersion, format A.B.C must be used"
                )
            self.major = splits[0]
            if len(splits) > 1:
                self.minor = splits[1]
            if len(splits) > 2:
                self.patch = splits[2]

        else:
            self.major = major_or_canonical_str
            self.minor = minor
            self.patch = patch

    def equals(self: "LibVersion", lib2: "LibVersion") -> bool:
        """
        Compare two LibVersions of format A.B.C , consisting of either Digits 0-9 or x .
        X equals to any other version, i.e.  3.X.X equals 3.5.1
        A Lib Version may have the pattern A.B or A
        :param lib2:
        :return:
        """
        if self.major == lib2.major:
            if self.minor == "x" or lib2.minor == "x":
                return True

            if self.minor == lib2.minor:
                if self.patch == "x" or lib2.patch == "x":
                    return True

                if self.patch == lib2.patch:
                    return True
        return False

    def is_other_greater(self: "LibVersion", other: "LibVersion") -> bool:
        # basically  checks self < other
        if self.as_str() == other.as_str():
            return False

            # Major Check
        if self.major < other.major:
            return True
        elif self.major > other.major:
            return False

        # Minor Check
        if self.minor < other.minor:
            return True
        elif self.minor > other.minor:
            return False

        # Patch Check,
        # Patch could be missing
        if self.patch and not other.patch:
            return True
        if not self.patch and other.patch:
            return False
        # Patch could also be a str
        if (
            isinstance(self.patch, int)
            and isinstance(other.patch, int)
            and self.patch < other.patch
        ):
            return True
        if (
            isinstance(self.patch, int)
            and isinstance(other.patch, int)
            and self.patch > other.patch
        ):
            return False

        # One lib must be rc, parse it
        if isinstance(self.patch, str) and "rc" in self.patch:
            self_patch, self_rc = self.patch.split("rc")
            self_patch = int(self_patch)
            self_rc = int(self_rc)
        else:
            self_patch = self.patch
            self_rc = None

        if isinstance(other.patch, str) and "rc" in self.patch:
            other_patch, other_rc = self.patch.split("rc")
            other_patch = int(other_patch)
            other_rc = int(other_rc)

        else:
            other_patch = other.patch
            other_rc = None

        if self_patch < other_patch:
            return True
        if self_patch > other_patch:
            return True

        # One may have an rc
        if not self_rc and other_rc:
            return True
        if self_rc and not other_rc:
            return False
        # Both have no rc
        if not self_rc and not other_rc:
            return False

        # Both have a int rc
        if self_rc < other_rc:
            return True
        if self_rc > other_rc:
            return True

        # Version is equal, not an upgrade
        return False

    def as_str(self) -> str:
        """Return LibVersion object as canonical str representation"""
        # We filter out all values != None soo version checks match up
        return ".".join(filter(lambda x: x, [self.major, self.minor, self.patch]))

    def __str__(self):
        return self.as_str()